@@ -1,14 +1,16 @@
 import collections
+import copy
 import csv
 import datetime
 import functools
 import json
 import logging
+import numbers
 import os
 import re
-from collections import OrderedDict
 from tempfile import NamedTemporaryFile
-from urllib.parse import urljoin, urlparse
+from urllib.parse import urljoin, urlparse, urlsplit
+from urllib.request import urlopen
 
 import jsonref
 import jsonschema.validators
@@ -16,17 +18,67 @@ import requests
 from cached_property import cached_property
 from flattentool import unflatten
 from jsonschema import FormatChecker, RefResolver
-from jsonschema._utils import extras_msg, find_additional_properties, uniq
-from jsonschema.compat import urlopen, urlsplit
-from jsonschema.exceptions import ValidationError
+from jsonschema._utils import (
+    ensure_list,
+    extras_msg,
+    find_additional_properties,
+    types_msg,
+    uniq,
+)
+from jsonschema.exceptions import UndefinedTypeCheck, ValidationError
 
 from .exceptions import cove_spreadsheet_conversion_error
 from .tools import decimal_default, get_request
 
+
+def type_validator(validator, types, instance, schema):
+    """
+    Replace the jsonschema type validator to use for-loop instead of slower
+    any() with generator expression.
+
+    https://github.com/OpenDataServices/lib-cove/pull/66
+
+    """
+    types = ensure_list(types)
+
+    for type in types:
+        if validator.is_type(instance, type):
+            break
+    else:
+        yield ValidationError(types_msg(instance, types))
+
+
+class TypeChecker:
+    def is_type(self, instance, type):
+        if type == "string":
+            return isinstance(instance, str)
+        if type == "array":
+            return isinstance(instance, list)
+        if type == "object":
+            return isinstance(instance, dict)
+        if type == "integer":
+            if isinstance(instance, bool):
+                return False
+            return isinstance(instance, int)
+        if type == "number":
+            if isinstance(instance, bool):
+                return False
+            return isinstance(instance, numbers.Number)
+        if type == "boolean":
+            return isinstance(instance, bool)
+        if type == "null":
+            return instance is None
+        raise UndefinedTypeCheck(type)
+
+
 # Because we will be changing items on this validator, it's important we take a copy!
 # Otherwise we could cause conflicts with other software in the same process.
 validator = jsonschema.validators.extend(
-    jsonschema.validators.Draft4Validator, validators={}
+    jsonschema.validators.Draft4Validator,
+    validators={
+        "type": type_validator,
+    },
+    type_checker=TypeChecker(),
 )
 
 uniqueItemsValidator = validator.VALIDATORS.pop("uniqueItems")
@@ -303,7 +355,7 @@ class SchemaJsonMixin:
 
     @property
     def _schema_obj(self):
-        return json.loads(self.schema_str, object_pairs_hook=OrderedDict)
+        return json.loads(self.schema_str)
 
     @property
     def _pkg_schema_obj(self):
@@ -344,13 +396,13 @@ def schema_dict_fields_generator(schema_dict):
                     continue
                 if "properties" in property_schema_dict:
                     for field in schema_dict_fields_generator(property_schema_dict):
-                        yield "/" + property_name + field
+                        yield f"/{property_name}{field}"
                 elif "items" in property_schema_dict:
                     for field in schema_dict_fields_generator(
                         property_schema_dict["items"]
                     ):
-                        yield "/" + property_name + field
-                yield "/" + property_name
+                        yield f"/{property_name}{field}"
+                yield f"/{property_name}"
     if "items" in schema_dict and isinstance(schema_dict["items"], dict):
         if "oneOf" in schema_dict["items"] and isinstance(
             schema_dict["items"]["oneOf"], list
@@ -429,7 +481,7 @@ def load_core_codelists(codelist_url, unique_files, config=None):
 @functools.lru_cache()
 def _deref_schema(schema_str, schema_host, cache=None):
     loader = CustomJsonrefLoader(schema_url=schema_host, cache=cache)
-    deref_obj = jsonref.loads(schema_str, loader=loader, object_pairs_hook=OrderedDict)
+    deref_obj = jsonref.loads(schema_str, loader=loader)
     # Force evaluation of jsonref.loads here
     repr(deref_obj)
     return deref_obj
@@ -503,9 +555,6 @@ def common_checks_context(
     )
 
     additional_fields_count = sum(item[2] for item in additional_fields)
-    additional_fields_all = get_additional_fields_info(
-        json_data, schema_fields, context, fields_regex=fields_regex
-    )
 
     context.update(
         {
@@ -575,7 +624,7 @@ def common_checks_context(
         }
     )
 
-    json_data_gen_paths = get_json_data_generic_paths(json_data)
+    json_data_gen_paths = get_json_data_generic_paths(json_data, generic_paths={})
     context["deprecated_fields"] = get_json_data_deprecated_fields(
         json_data_gen_paths, schema_obj
     )
@@ -626,8 +675,8 @@ def get_additional_codelist_values(schema_obj, json_data):
                     if schema_obj.extended_codelist_urls.get(codelist):
                         codelist_url = schema_obj.extended_codelist_urls[codelist][-1]
 
-                    codelistadd = "+" + codelist
-                    codelistsub = "-" + codelist
+                    codelistadd = f"+{codelist}"
+                    codelistsub = f"-{codelist}"
                     for codelist_key in schema_obj.extended_codelist_urls.keys():
                         if codelist_key == codelistadd:
                             for amended_codelist in schema_obj.extended_codelist_urls[
@@ -663,7 +712,7 @@ def get_additional_codelist_values(schema_obj, json_data):
 def get_additional_fields_info(json_data, schema_fields, context, fields_regex=False):
     fields_present = get_fields_present_with_examples(json_data)
 
-    additional_fields = collections.OrderedDict()
+    additional_fields = {}
     root_additional_fields = set()
 
     for field, field_info in fields_present.items():
@@ -682,7 +731,7 @@ def get_additional_fields_info(json_data, schema_fields, context, fields_regex=F
                 break
         else:
             field_info["root_additional_field"] = True
-            field_info["additional_field_descendance"] = collections.OrderedDict()
+            field_info["additional_field_descendance"] = {}
             root_additional_fields.add(field)
 
         field_info["path"] = "/".join(field.split("/")[:-1])
@@ -821,7 +870,7 @@ def get_schema_validation_errors(
                 else:
                     parent_name = e.path[-1]
 
-            heading = heading_src_map.get(path_no_number + "/" + e.message)
+            heading = heading_src_map.get(f"{path_no_number}/{e.message}")
             if heading:
                 field_name = heading[0][1]
                 value["header"] = heading[0][1]
@@ -862,34 +911,29 @@ def get_schema_validation_errors(
         if header_extra is None:
             header_extra = header
 
-        unique_validator_key = OrderedDict(
-            [
-                ("message", message),
-                ("validator", e.validator),
-                ("assumption", e.assumption if hasattr(e, "assumption") else None),
-                # Don't pass this value for 'enum' and 'required' validators,
-                # because it is not needed, and it will mean less grouping, which
-                # we don't want.
-                (
-                    "validator_value",
-                    e.validator_value
-                    if e.validator not in ["enum", "required"]
-                    else None,
-                ),
-                ("message_type", validator_type),
-                ("path_no_number", path_no_number),
-                ("header", header),
-                ("header_extra", header_extra),
-                ("null_clause", null_clause),
-                ("error_id", e.error_id if hasattr(e, "error_id") else None),
-                ("exclusiveMinimum", e.schema.get("exclusiveMinimum")),
-                ("exclusiveMaximum", e.schema.get("exclusiveMaximum")),
-                ("extras", getattr(e, "extras", None)),
-                ("each", getattr(e, "each", None)),
-                ("property", getattr(e, "property", None)),
-                ("reprs", getattr(e, "reprs", None)),
-            ]
-        )
+        unique_validator_key = {
+            "message": message,
+            "validator": e.validator,
+            "assumption": e.assumption if hasattr(e, "assumption") else None,
+            # Don't pass this value for 'enum' and 'required' validators,
+            # because it is not needed, and it will mean less grouping, which
+            # we don't want.
+            "validator_value": e.validator_value
+            if e.validator not in ["enum", "required"]
+            else None,
+            "message_type": validator_type,
+            "path_no_number": path_no_number,
+            "header": header,
+            "header_extra": header_extra,
+            "null_clause": null_clause,
+            "error_id": e.error_id if hasattr(e, "error_id") else None,
+            "exclusiveMinimum": e.schema.get("exclusiveMinimum"),
+            "exclusiveMaximum": e.schema.get("exclusiveMaximum"),
+            "extras": getattr(e, "extras", None),
+            "each": getattr(e, "each", None),
+            "property": getattr(e, "property", None),
+            "reprs": getattr(e, "reprs", None),
+        }
         if instance is not None:
             unique_validator_key["instance"] = instance
         validation_errors[
@@ -898,7 +942,7 @@ def get_schema_validation_errors(
     return dict(validation_errors)
 
 
-def get_json_data_generic_paths(json_data, path=(), generic_paths=None):
+def get_json_data_generic_paths(json_data, generic_paths, path=(), generic_key=()):
     """Transform json data into a dictionary with keys made of json paths.
 
     Key are json paths (as tuples). Values are dictionaries with keys including specific
@@ -932,28 +976,26 @@ def get_json_data_generic_paths(json_data, path=(), generic_paths=None):
         ('c', 'cb'): {('c', 2, 'cb'): 'cb'}
     }
     """
-    if generic_paths is None:
-        generic_paths = {}
-
-    if isinstance(json_data, dict):
-        iterable = list(json_data.items())
-        if not iterable:
-            generic_paths[path] = {}
+    if type(json_data) is list:
+        is_dict = False
+        iterable = enumerate(json_data)
+        new_generic_key = generic_key
     else:
-        iterable = list(enumerate(json_data))
-        if not iterable:
-            generic_paths[path] = []
+        is_dict = True
+        iterable = json_data.items()
 
     for key, value in iterable:
-        generic_key = tuple(i for i in path + (key,) if type(i) != int)
+        new_path = path + (key,)
+        if is_dict:
+            new_generic_key = generic_key + (key,)
 
-        if generic_paths.get(generic_key):
-            generic_paths[generic_key][path + (key,)] = value
+        if new_generic_key in generic_paths:
+            generic_paths[new_generic_key][new_path] = value
         else:
-            generic_paths[generic_key] = {path + (key,): value}
+            generic_paths[new_generic_key] = {new_path: value}
 
         if isinstance(value, (dict, list)):
-            get_json_data_generic_paths(value, path + (key,), generic_paths)
+            get_json_data_generic_paths(value, generic_paths, new_path, new_generic_key)
 
     return generic_paths
 
@@ -963,10 +1005,10 @@ def get_json_data_deprecated_fields(json_data_paths, schema_obj):
     deprecated_json_data_paths = [
         path for path in deprecated_schema_paths if path[0] in json_data_paths
     ]
-    # Generate an OrderedDict sorted by deprecated field names (keys) mapping
+    # Generate a dict sorted by deprecated field names (keys) mapping
     # to a unordered tuple of tuples:
     # {deprecated_field: ((path, path... ), (version, description))}
-    deprecated_fields = OrderedDict()
+    deprecated_fields = {}
     for generic_path in sorted(deprecated_json_data_paths, key=lambda tup: tup[0][-1]):
         deprecated_fields[generic_path[0][-1]] = tuple()
 
@@ -985,7 +1027,7 @@ def get_json_data_deprecated_fields(json_data_paths, schema_obj):
             )
 
     # Order the path tuples in values for deprecated_fields.
-    deprecated_fields_output = OrderedDict()
+    deprecated_fields_output = {}
     for field, paths in deprecated_fields.items():
         sorted_paths = tuple(sorted(paths[0]))
 
@@ -1045,11 +1087,12 @@ def _generate_data_path(json_data, path=()):
 
 
 def get_fields_present_with_examples(*args, **kwargs):
-    counter = collections.OrderedDict()
+    counter = {}
     for key, value in fields_present_generator(*args, **kwargs):
         if key not in counter:
-            counter[key] = {"count": 0, "examples": []}
-        counter[key]["count"] += 1
+            counter[key] = {"count": 1, "examples": []}
+        else:
+            counter[key]["count"] += 1
         if len(counter[key]["examples"]) < 3:
             if not isinstance(value, (list, dict)):
                 counter[key]["examples"].append(value)
@@ -1096,7 +1139,7 @@ class CustomRefResolver(RefResolver):
 
             if scheme in self.handlers:
                 result = self.handlers[scheme](uri)
-            elif scheme in [u"http", u"https"]:
+            elif scheme in ["http", "https"]:
                 # Requests has support for detecting the correct encoding of
                 # json over http
                 result = get_request(uri, config=self.config).json()
@@ -1233,16 +1276,10 @@ def _get_schema_non_required_ids(
 def fields_present_generator(json_data, prefix=""):
     if isinstance(json_data, dict):
         for key, value in json_data.items():
-            if isinstance(value, list):
-                yield prefix + "/" + key, value
-                for item in value:
-                    if isinstance(item, dict):
-                        yield from fields_present_generator(item, prefix + "/" + key)
-            elif isinstance(value, dict):
-                yield prefix + "/" + key, value
-                yield from fields_present_generator(value, prefix + "/" + key)
-            else:
-                yield prefix + "/" + key, value
+            new_key = f"{prefix}/{key}"
+            yield new_key, value
+            if isinstance(value, (dict, list)):
+                yield from fields_present_generator(value, new_key)
     elif isinstance(json_data, list):
         for item in json_data:
             if isinstance(item, dict):
@@ -1350,3 +1387,112 @@ def get_orgids_prefixes(orgids_url=None):
         os.rename(tmp.name, local_org_ids_file)
     # Return either the original file data, if it was found to be fresh, or the new data, if we were able to retrieve it.
     return [org_list["code"] for org_list in org_id_file_contents["lists"]]
+
+
+def add_field_coverage(schema_dict, json_data):
+    """
+    Takes a schema dict and adds non-zero coverage counts of successes and
+    checks, based on what is in the json data.
+
+    Doesn't support all possible json schemas e.g. anyOf, allOf, oneOf and an
+    array as the value for items are not supported, and will be ignored, along
+    with all their children.
+
+    """
+    if not isinstance(schema_dict, dict):
+        return schema_dict
+
+    schema_properties = schema_dict.get("properties")
+    if isinstance(schema_properties, dict) and isinstance(json_data, dict):
+        for schema_property, sub_schema_obj in schema_properties.items():
+            if "coverage" not in sub_schema_obj:
+                sub_schema_obj["coverage"] = {}
+            sub_schema_obj["coverage"]["checks"] = (
+                sub_schema_obj.get("coverage", {}).get("checks", 0) + 1
+            )
+            if json_data.get(schema_property):
+                sub_schema_obj["coverage"]["successes"] = (
+                    sub_schema_obj.get("coverage", {}).get("successes", 0) + 1
+                )
+            add_field_coverage(sub_schema_obj, json_data.get(schema_property))
+
+    schema_items = schema_dict.get("items")
+    if isinstance(schema_items, dict) and isinstance(json_data, list):
+        for json_data_item in json_data:
+            add_field_coverage(schema_items, json_data_item)
+    return schema_dict
+
+
+def add_field_coverage_percentages(schema_dict):
+    """
+    Takes the output of add_field_coverage and adds percentages, and also zero
+    counts of successes and checks.
+
+    """
+    if not isinstance(schema_dict, dict):
+        return schema_dict
+
+    schema_properties = schema_dict.get("properties")
+    if isinstance(schema_properties, dict):
+        for schema_property, sub_schema_obj in schema_properties.items():
+            if "coverage" not in sub_schema_obj:
+                sub_schema_obj["coverage"] = {}
+            if "checks" not in sub_schema_obj["coverage"]:
+                sub_schema_obj["coverage"]["checks"] = 0
+            if "successes" not in sub_schema_obj["coverage"]:
+                sub_schema_obj["coverage"]["successes"] = 0
+            if (
+                sub_schema_obj["coverage"]["checks"] == 0
+                and sub_schema_obj["coverage"]["successes"] == 0
+            ):
+                sub_schema_obj["coverage"]["percentage"] = 0
+            else:
+                sub_schema_obj["coverage"]["percentage"] = int(
+                    sub_schema_obj["coverage"]["successes"]
+                    / sub_schema_obj["coverage"]["checks"]
+                    * 100
+                )
+            add_field_coverage_percentages(sub_schema_obj)
+
+    add_field_coverage_percentages(schema_dict.get("items"))
+    return schema_dict
+
+
+def dict_copy(dict_in):
+    """
+    Make a copy of a dict, and any other dicts nested inside it.
+
+    We use this instead of copy.deepcopy because that keeps a memo, and creates
+    only one new object for objects that are the same.
+
+    This doesn't deal with the use of other objects for nesting, e.g. lists,
+    because the code using the resulting structure only descends into dicts.
+
+    """
+    dict_out = copy.copy(dict_in)
+    for key, value in dict_out.items():
+        if isinstance(value, dict):
+            dict_out[key] = dict_copy(value)
+    return dict_out
+
+
+def get_field_coverage(schema_obj, json_data_list):
+    """
+    Returns a copy of the schema with coverage counts of successes, checks and
+    percentages annoated.
+
+    json_data_list is the main list of objects in the json data, as we use the
+    main schema, not the package schema.
+    e.g. for OC4IDS we call this function with:
+    get_field_coverage(schema_obj, json_data.get("projects")
+
+    """
+    # Need to call dict_copy here because jsonref returns the same dict when
+    # there's mutliple refs to the same place.
+    schema_dict = dict_copy(schema_obj.get_schema_obj(deref=True))
+    if not isinstance(json_data_list, list):
+        return {}
+    for json_data_item in json_data_list:
+        add_field_coverage(schema_dict, json_data_item)
+    add_field_coverage_percentages(schema_dict)
+    return schema_dict
